import mimetypes
from typing import Optional, Sequence
import cgi

from requests import Response
from requests.structures import CaseInsensitiveDict
from requests_toolbelt.multipart.decoder import MultipartDecoder

from rets.errors import RetsApiError, RetsResponseError
from rets.http.data import Object
from rets.http.parsers.parse import DEFAULT_ENCODING, ResponseLike, parse_xml


def parse_object(response: Response) -> Sequence[Object]:
    """
    Parse the response from a GetObject transaction. If there are multiple
    objects to be returned then the response should be a multipart response.
    The headers of the response (or each part in the multipart response)
    contains the metadata for the object, including the location if requested.
    The body of the response should contain the binary content of the object,
    an XML document specifying a transaction status code, or left empty.
    """
    content_type = response.headers.get('content-type')

    if content_type and 'multipart/parallel' in content_type:
        return _parse_multipart(response)

    object_ = _parse_body_part(response)
    return (object_,) if object_ is not None else ()


def _parse_multipart(response: ResponseLike) -> Sequence[Object]:
    """
    RFC 2045 describes the format of an Internet message body containing a MIME message. The
    body contains one or more body parts, each preceded by a boundary delimiter line, and the
    last one followed by a closing boundary delimiter line. After its boundary delimiter line,
    each body part then consists of a header area, a blank line, and a body area.

    HTTP/1.1 200 OK
    Server: Apache/2.0.13
    Date: Fri, 22 OCT 2004 12:03:38 GMT
    Cache-Control: private
    RETS-Version: RETS/1.7.2
    MIME-Version: 1.0
    Content-Type: multipart/parallel; boundary="simple boundary"

    --simple boundary
    Content-Type: image/jpeg
    Content-ID: 123456
    Object-ID: 1

    <binary data>

    --simple boundary
    Content-Type: text/xml
    Content-ID: 123457
    Object-ID: 1

    <RETS ReplyCode="20403" ReplyText="There is no listing with that ListingID"/>

    --simple boundary--
    """
    encoding = response.encoding or DEFAULT_ENCODING
    multipart = MultipartDecoder.from_response(response, encoding)
    # We need to decode the headers because MultipartDecoder returns bytes keys and values,
    # while requests.Response.headers uses str keys and values.
    for part in multipart.parts:
        part.headers = _decode_headers(part.headers, encoding)

    objects = (_parse_body_part(part) for part in multipart.parts)
    return tuple(object_ for object_ in objects if object_ is not None)


def _parse_body_part(part: ResponseLike) -> Optional[Object]:
    headers = part.headers

    content_id = headers.get('content-id')
    object_id = headers.get('object-id')
    preferred = 'preferred' in headers
    description = headers.get('content-description')
    location = headers.get('location')
    content_type = headers.get('content-type')
    mime_type = _parse_mime_type(content_type) if content_type else None

    # Check XML responses first, it may contain an error description.
    if mime_type == 'text/xml':
        try:
            parse_xml(part)
        except RetsApiError as e:
            if e.reply_code == 20403:  # No object found
                return None
            raise

    # All RETS responses _must_ have `Content-ID` and `Object-ID` headers.
    if not content_id or not object_id:
        raise RetsResponseError(part.content, part.headers)

    # Respond with `Location` header redirect.
    if location:
        return Object(
            mime_type=_guess_mime_type(location) or mime_type,
            content_id=content_id,
            description=description,
            object_id=object_id,
            url=location,
            preferred=preferred,
            data=None,
        )

    # Check the `Content-Type` header exists for object responses.
    if mime_type is None or mime_type == 'text/html':
        raise RetsResponseError(part.content, part.headers)

    return Object(
        mime_type=mime_type,
        content_id=content_id,
        description=description,
        object_id=object_id,
        url=None,
        preferred=preferred,
        data=part.content or None,
    )


def _guess_mime_type(location: str) -> Optional[str]:
    mime_type, _ = mimetypes.guess_type(location)
    return mime_type


def _parse_mime_type(content_type: str) -> Optional[str]:
    # Parse mime type from content-type header, e.g. 'image/jpeg;charset=US-ASCII' -> 'image/jpeg'
    mime_type, _ = cgi.parse_header(content_type)
    return mime_type or None


def _decode_headers(headers: CaseInsensitiveDict, encoding: str) -> CaseInsensitiveDict:
    return CaseInsensitiveDict({
        k.decode(encoding): v.decode(encoding)
        for k, v in headers.items()
    })



from typing import FrozenSet, Mapping, Sequence, Union

from rets.client.decoder import RecordDecoder
from rets.client.record import Record
from rets.client.utils import get_metadata_data
from rets.errors import RetsClientError
from rets.http import RetsHttpClient, SearchResult


class ResourceClass:

    def __init__(self, resource, metadata: dict, http_client: RetsHttpClient):
        self.resource = resource
        self._http = http_client
        self._metadata = metadata
        self._table = metadata.get('_table')
        self._fields = None

    @property
    def name(self) -> str:
        return self._metadata['ClassName']

    @property
    def has_key_index(self) -> bool:
        return self._metadata.get('HasKeyIndex') == '1'

    @property
    def metadata(self) -> dict:
        metadata = dict(self._metadata)
        if self._table:
            metadata['_table'] = self._table
        return metadata

    @property
    def table(self) -> Sequence[dict]:
        if self._table is None:
            self._table = tuple(get_metadata_data(self._http, 'table', resource=self.resource.name, class_=self.name))
        return self._table

    @property
    def fields(self) -> FrozenSet[str]:
        if self._fields is None:
            self._fields = frozenset(field['SystemName'] for field in self.table)
        return self._fields

    def search(self,
               query: Union[str, Mapping[str, str]],
               fields: Sequence[str] = None,
               parse: bool = True,
               include_tz: bool = False,
               **kwargs) -> SearchResult:
        query = self._validate_query(query)
        if fields:
            fields = self._validate_fields(fields)

        result = self._http.search(
            resource=self.resource.name,
            class_=self.name,
            query=query,
            select=fields,
            **kwargs,
        )

        if parse:
            decoder = RecordDecoder(self.table, include_tz)
            rows = decoder.decode(result.data)
        else:
            rows = result.data

        return SearchResult(
            count=result.count,
            max_rows=result.max_rows,
            data=tuple(Record(self, row) for row in rows),
        )

    def _validate_query(self, query: Union[str, Mapping[str, str]]) -> str:
        if isinstance(query, str):
            return query
        self._assert_fields(query)
        return ','.join('(%s=%s)' % (field, value) for field, value in query.items())

    def _validate_fields(self, fields: Sequence[str]) -> str:
        self._assert_fields(fields)
        return ','.join(fields)

    def _assert_fields(self, fields: Sequence[str]) -> None:
        permissible = self.fields
        invalid = tuple(f for f in fields if f not in permissible)
        if invalid:
            raise RetsClientError('invalid fields %s' % ','.join(invalid))

    def __repr__(self) -> str:
        return '<Class: %s:%s>' % (self.resource.name, self.name)



from rets.http.parsers.parse import (
    parse_capability_urls,
    parse_metadata,
    parse_search,
    parse_system,
)
from rets.http.parsers.parse_object import parse_object

__all__ = [
    'parse_capability_urls',
    'parse_metadata',
    'parse_object',
    'parse_search',
    'parse_system',
]



from collections import OrderedDict
from itertools import zip_longest
from typing import Iterable, Sequence, Tuple, Union
from lxml import etree

from requests import Response
from requests_toolbelt.multipart.decoder import BodyPart

from rets.errors import RetsParseError, RetsApiError, RetsResponseError
from rets.http.data import Metadata, SearchResult, SystemMetadata

DEFAULT_ENCODING = 'utf-8'

ResponseLike = Union[Response, BodyPart]


def parse_xml(response: ResponseLike) -> etree.Element:
    encoding = response.encoding or DEFAULT_ENCODING
    root = etree.fromstring(response.content.decode(encoding), parser=etree.XMLParser(recover=True))

    if root is None:
        raise RetsResponseError(response.content, response.headers)

    reply_code, reply_text = _parse_rets_status(root)
    if reply_code and reply_text != "Operation Successful":
        raise RetsApiError(reply_code, reply_text, response.content)

    return root


def parse_capability_urls(response: Response) -> dict:
    """
    Parses the list of capability URLs from the response of a successful Login transaction.

    The capability url list is the set of functions or URLs to which the Login grants access.
    A capability consists of a key and a URL. The list returned from the server in the login
    reply must include URLs for Search, Login, and GetMetadata, and optionally may include
    URLs for Action, ChangePassword, GetObject, LoginComplete, Logout, ServerInformation,
    and Update.

    <RETS ReplyCode="0" ReplyText="Success">
        <RETS-RESPONSE>
            MemberName=member_name
            User=user_id,user_level,user_class,agent_code
            Broker=RETSOFFIC
            MetadataVersion=01.09.02991
            MetadataTimestamp=2016-11-24T05:24:06Z
            MinMetadataTimestamp=2016-11-24T05:24:06Z
            Login=/rets2_1/Login
            Search=/rets2_1/Search
            GetMetadata=/rets2_1/GetMetadata
            GetObject=/rets2_1/GetObject
            Logout=/rets2_1/Logout
        </RETS-RESPONSE>
    </RETS>
    """
    elem = parse_xml(response)
    response_elem = elem.find('RETS-RESPONSE')
    if response_elem is None:
        return {}
    raw_arguments = response_elem.text.strip().split('\n')
    return dict((s.strip() for s in arg.split('=', 1)) for arg in raw_arguments)


def parse_metadata(response: Response) -> Sequence[Metadata]:
    """
    Parse the information from a GetMetadata transaction.

    <RETS ReplyCode="0" ReplyText="Success">
        <METADATA-RESOURCE Date="2016-11-24T05:24:06Z" Version="01.09.02991">
            <COLUMNS>	ResourceID	StandardName	</COLUMNS>
            <DATA>	ActiveAgent	ActiveAgent	</DATA>
            <DATA>	Office	Office	</DATA>
            <DATA>	OpenHouse	OpenHouse	</DATA>
            <DATA>	Property	Property	</DATA>
            <DATA>	RentalSchedule	RentalSchedule	</DATA>
        </METADATA-RESOURCE>
    </RETS>
    """
    elem = parse_xml(response)
    metadata_elems = [e for e in elem.findall('*') if e.tag.startswith('METADATA-')]
    if metadata_elems is None:
        return ()

    def parse_metadata_elem(elem: etree.Element) -> Metadata:
        """ Parses a single <METADATA-X> element """
        return Metadata(
            type_=elem.tag.split('-', 1)[1],
            resource=elem.get('Resource'),
            class_=elem.get('Class'),
            data=tuple(_parse_data(elem)),
        )

    return tuple(parse_metadata_elem(metadata_elem) for metadata_elem in metadata_elems)


def parse_system(response: Response) -> SystemMetadata:
    """
    Parse the server system information from a SYSTEM GetMetadata transaction.

    <RETS ReplyCode="0" ReplyText="Success">
        <METADATA-SYSTEM Date="2016-11-24T05:24:06Z" Version="01.09.02991">
            <SYSTEM SystemDescription="ARMLS" SystemID="az" TimeZoneOffset="-06:00"/>
            <COMMENTS/>
        </METADATA-SYSTEM>
    </RETS>
    """
    elem = parse_xml(response)
    metadata_system_elem = _find_or_raise(elem, 'METADATA-SYSTEM')
    system_elem = _find_or_raise(metadata_system_elem, 'SYSTEM')
    comments_elem = metadata_system_elem.find('COMMENTS')
    return SystemMetadata(
        system_id=system_elem.get('SystemID'),
        system_description=system_elem.get('SystemDescription'),
        system_date=metadata_system_elem.get('Date'),
        system_version=metadata_system_elem.get('Version'),

        # Optional fields
        time_zone_offset=system_elem.get('TimeZoneOffset'),
        comments=comments_elem and (comments_elem.text or None),
    )


def parse_search(response: Response) -> SearchResult:
    try:
        elem = parse_xml(response)
    except RetsApiError as e:
        if e.reply_code == 20201:  # No records found
            return SearchResult(0, False, ())
        raise

    count_elem = elem.find('COUNT')
    if count_elem is not None:
        count = int(count_elem.get('Records'))
    else:
        count = None

    try:
        data = tuple(_parse_data(elem))
    except RetsParseError:
        data = None

    return SearchResult(
        count=count,
        # python xml.etree.ElementTree.Element objects are always considered false-y
        max_rows=elem.find('MAXROWS') is not None,
        data=data,
    )


def _parse_rets_status(root: etree.Element) -> Tuple[int, str]:
    """
    If RETS-STATUS exists, the client must use this instead
    of the status from the body-start-line
    """
    rets_status = root.find('RETS-STATUS')
    elem = rets_status if rets_status is not None else root
    return int(elem.get('ReplyCode')), elem.get('ReplyText')


def _parse_data(elem: etree.Element) -> Iterable[dict]:
    """
    Parses a generic container element enclosing a single COLUMNS and multiple DATA elems, and
    returns a generator of dicts with keys given by the COLUMNS elem and values given by each
    DATA elem. The container elem may optionally contain a DELIMITER elem to define the delimiter
    used, otherwise a default of '\t' is assumed.

    <RETS ReplyCode="0" ReplyText="Success">
        <DELIMITER value="09"/>
        <COLUMNS>	LIST_87	LIST_105	LIST_1	</COLUMNS>
        <DATA>	2016-12-01T00:08:10	5489015	20160824051756837742000000	</DATA>
        <DATA>	2016-12-01T00:10:02	5497756	20160915055426038684000000	</DATA>
        <DATA>	2016-12-01T00:10:26	5528935	20161123230848928777000000	</DATA>
        <DATA>	2016-12-01T00:10:52	5528955	20161123234916869427000000	</DATA>
        <DATA>	2016-12-01T00:14:31	5530021	20161127221848669500000000	</DATA>
    </RETS>
    """
    delimiter = _parse_delimiter(elem)

    columns_elem = _find_or_raise(elem, 'COLUMNS')
    columns = _parse_data_line(columns_elem, delimiter)

    data_elems = elem.findall('DATA')

    return (OrderedDict(zip_longest(columns, _parse_data_line(data, delimiter)))
            for data in data_elems)


def _find_or_raise(elem: etree.Element, child_elem_name: str) -> etree.Element:
    child = elem.find(child_elem_name)
    if child is None:
        raise RetsParseError('Missing %s element' % child_elem_name)
    return child


def _parse_data_line(elem: etree.Element, delimiter: str = '\t') -> Sequence[str]:
    # DATA elems using the COMPACT format and COLUMN elems all start and end with delimiters
    return elem.text.split(delimiter)[1:-1]


def _parse_delimiter(elem: etree.Element) -> str:
    delimiter_elem = elem.find('DELIMITER')
    if delimiter_elem is None:
        return '\t'
    return chr(int(delimiter_elem.get('value')))



from rets.http.client import RetsHttpClient
from rets.http.data import Metadata, Object, SearchResult, SystemMetadata

__all__ = [
    'Metadata',
    'Object',
    'RetsHttpClient',
    'SearchResult',
    'SystemMetadata',
]



from rets.client import RetsClient
from rets.http.client import RetsHttpClient
from rets.http.data import Metadata, Object, SearchResult, SystemMetadata

__title__ = 'rets'
__version__ = '0.4.5'
__author__ = 'Martin Liu <martin@opendoor.com>'
__license__ = 'MIT License'

__all__ = [
    'RetsClient',
    'RetsHttpClient',
    'Metadata',
    'Object',
    'SearchResult',
    'SystemMetadata',
]



from collections import namedtuple

Metadata = namedtuple('Metadata', (
    'type_',
    'resource',
    'class_',
    'data',
))

Object = namedtuple('Object', (
    'mime_type',
    'content_id',
    'description',
    'object_id',
    'url',
    'preferred',
    'data',
))

SearchResult = namedtuple('SearchResult', (
    'count',
    'max_rows',
    'data',
))

SystemMetadata = namedtuple('SystemMetadata', (
    'system_id',
    'system_description',
    'system_date',
    'system_version',
    'time_zone_offset',
    'comments',
))



from rets.http.client import RetsHttpClient


def get_metadata_data(http_client: RetsHttpClient, type_: str, **kwargs):
    metadata_structs = http_client.get_metadata(type_, **kwargs)
    if metadata_structs:
        return metadata_structs[0].data
    return ()



from hashlib import md5
from typing import Any, Mapping, Sequence, Union
from urllib.parse import urljoin, urlsplit, urlunsplit, urlencode

import requests
from requests import Response
from requests.auth import AuthBase, HTTPBasicAuth, HTTPDigestAuth

from rets.http.parsers import (
    parse_capability_urls,
    parse_metadata,
    parse_object,
    parse_search,
    parse_system,
)
from rets.http.data import Object, Metadata, SearchResult, SystemMetadata
from rets.errors import RetsApiError, RetsClientError


class RetsHttpClient:

    def __init__(self,
                 login_url: str,
                 username: str = None,
                 password: str = None,
                 auth_type: str = 'digest',
                 user_agent: str = 'rets-python/0.3',
                 user_agent_password: str = '',
                 rets_version: str = '1.7.2',
                 capability_urls: str = None,
                 cookie_dict: dict = None,
                 use_get_method: bool = False,
                 ):
        self._user_agent = user_agent
        self._user_agent_password = user_agent_password
        self._rets_version = rets_version
        self._use_get_method = use_get_method

        splits = urlsplit(login_url)
        self._base_url = urlunsplit((splits.scheme, splits.netloc, '', '', ''))
        self._capabilities = capability_urls or {
            'Login': splits.path,
        }

        # Authenticate using either the user agent auth header and (basic or digest) HTTP auth.
        # SFARMLS (San Francisco) uses both methods together.
        if username and password:
            self._http_auth = _get_http_auth(username, password, auth_type)
        else:
            self._http_auth = None

        # we use a session to keep track of cookies that are required for certain MLSes
        self._session = requests.Session()

        # The user may provide an optional cookie_dict argument, which will be used on first login.
        # When sending cookies (with a session_id) to the login url, the same cookie (session_id)
        # is returned, which (most likely) means no additional login is created.
        if cookie_dict:
            for name, value in cookie_dict.items():
                self._session.cookies.set(name, value=value)

        # this session id is part of the rets standard for use with a user agent password
        self._rets_session_id = ''

    @property
    def user_agent(self) -> str:
        """
        This header field contains information about the user agent originating the request.
        This is for statistical purposes, the tracing of protocol violations, and automated
        recognition of user agents for the sake of tailoring responses to avoid particular user
        agent limitations, as well as providing enhanced capabilities to some user-agents. All
        client requests MUST include this field. This is a standard HTTP header field as defined
        in RFC 2616.
        """
        return self._user_agent

    @property
    def rets_version(self) -> str:
        """
        The client MUST send the RETS-Version. The convention used is a "<major>.<minor>.<release>"
        numbering scheme similar to the HTTP Version in Section 3.1 of RFC 2616. The version of a
        RETS message is indicated by a RETS-Version field in the header of the message.
        """
        return 'RETS/' + self._rets_version

    @property
    def capability_urls(self) -> dict:
        return self._capabilities

    @property
    def cookie_dict(self) -> dict:
        """Keeps the last value in case of duplicate keys."""
        cookie_d = {}
        for k, v in self._session.cookies.iteritems():
            cookie_d[k] = v
        return cookie_d

    def login(self) -> dict:
        response = self._http_request(self._url_for('Login'))
        self._capabilities = parse_capability_urls(response)
        return self._capabilities

    def logout(self) -> None:
        self._http_request(self._url_for('Logout'))
        self._session = None

    def get_system_metadata(self) -> SystemMetadata:
        return parse_system(self._get_metadata('system'))

    def get_metadata(self,
                     type_: str,
                     resource: str = None,
                     class_: str = None,
                     metadata_id: str = '0',
                     ) -> Sequence[Metadata]:
        if resource:
            id_ = ':'.join(filter(None, [resource, class_]))
        else:
            id_ = metadata_id

        try:
            return parse_metadata(self._get_metadata(type_, id_))
        except RetsApiError as e:
            if e.reply_code in (20502, 20503):  # No metadata exists.
                return ()
            raise

    def _get_metadata(self, type_: str, metadata_id: str = '0') -> Response:
        """
        :param type_: The type of metadata being requested. The Type MUST begin with METADATA and
            may be one of the defined metadata types (see Section 11).

        :param metadata_id: If the last metadata_id is 0 (zero), then the request is for all Type
            metadata contained within that level; if the last metadata_id is '*', then the request
            is for all Type metadata contained within that level and all metadata Types contained
            within the requested Type. This means that for a metadata-id of METADATA-SYSTEM, for
            example, the server is expected to return all metadata.

            Note: The metadata_id for METADATA-SYSTEM and METADATA-RESOURCE must be 0 or *.
        """
        payload = {
            'Type': 'METADATA-' + type_.upper(),
            'ID': metadata_id,
            'Format': 'COMPACT',
        }
        return self._http_request(self._url_for('GetMetadata'), payload=payload)

    def search(self,
               resource: str,
               class_: str,
               query: str,
               select: str = None,
               count: int = 1,
               limit: int = None,
               offset: int = 1,
               restricted_indicator: str = None,
               standard_names: bool = False,
               query_type: str = 'DMQL2',
               format_: str = 'COMPACT-DECODED',
               ) -> SearchResult:
        """
        The Search transaction requests that the server search one or more searchable databases
        and return the list of qualifying records. The body of the response contains the records
        matching the query, presented in the requested format.

        :param resource: The type of search to perform as discussed in Section 7.1 and defined
            in the Metadata (see section 11.2.2).

        :param class_: This parameter is set to a value that represents the class of data within
            the resource, taken from the Class metadata (section 11.3.1). If the resource has no
            classes, the class_ parameter will be ignored by the server and may be omitted by the
            client. If the client does include the class_ parameter for a classless search, the
            value should be the same as the resource in order to insure forward compatibility.

        :param query:

        :param count: If this argument is set to one '1', then a record-count is returned in the
            response in addition to the data. Note that on some servers this will cause the
            search to take longer since the count must be returned before any records are
            received. If this entry is set to two '2' then only a record-count is returned; no
            data is returned, but all matches are counted regardless of any Offset or Limit
            parameter. If the Count argument is not present or set to zero '0' then no record
            count is returned.

        :param limit:

        :param offset:

        :param restricted_indicator:

        :param standard_names: Queries may use either standard names or system names in the query
            (Section 7.7). If the client chooses to use standard names, it must indicate this
            using the standard_names argument. If this entry is set to '0' or is not present the
            field names passed in the search are the SystemNames, as defined in the metadata. If
            this entry is set to '1' then the StandardNames are used for the field names passed
            in the search. The StandardName designation applies to all names used in the query:
            SearchType, Class, Query and Select.

        :param format_: 'COMPACT' means a field list <COLUMNS> followed by a delimited set of the
            data fields <DATA>. 'COMPACT-DECODED' is the same as COMPACT except the data for any
            field with an interpretation of Lookup, LookupMulti, LookupBitString or LookupBitMask,
            is returned in a fully-decoded format using the LongValue. See Section 13 for more
            information on the COMPACT formats and section 11.4.3 for more information on the
            Lookup types. 'STANDARD-XML' means an XML presentation of the data in the format
            defined by the RETS Data XML DTD. Servers MUST support all formats. If the format is
            not specified, the server MUST return STANDARD-XML.
        """
        raw_payload = {
            'SearchType': resource,
            'Class': class_,
            'Query': query,
            'QueryType': query_type,
            'Select': select,
            'Count': count,
            'Limit': limit or 'NONE',
            'Offset': offset,
            'RestrictedIndicator': restricted_indicator,
            'StandardNames': int(standard_names),
            'Format': format_,
        }
        # None values indicate that the argument should be omitted from the request
        payload = {k: v for k, v in raw_payload.items() if v is not None}

        response = self._http_request(self._url_for('Search'), payload=payload)
        return parse_search(response)

    def get_object(self,
                   resource: str,
                   object_type: str,
                   resource_keys: Union[str, Mapping[str, Any], Sequence[str]],
                   media_types: Union[str, Sequence[str]] = '*/*',
                   location: bool = False,
                   ) -> Sequence[Object]:
        """
        The GetObject transaction is used to retrieve structured information related to known
        system entities. It can be used to retrieve multimedia files and other key-related
        information. Objects requested and returned from this transaction are requested and
        returned as MIME media types. The message body for successful retrievals contains only
        the objects in the specified MIME media type. Error responses follow the normal response
        format (section 3.9).

        :param resource: A resource defined in the metadata dictionary (see Section 11.2.2). The
            resource from which the object should be retrieved is specified by this entry. For
            more information see 5.9. The resource must be a resource defined in the metadata
            (section 11.4.1).

        :param object_type: The object type as defined in the metadata (see Section 11.4.1). The
            grouping category to which the object belongs. The type must be an ObjectType defined
            in the Object metadata for this Resource. For more information see section 11.4.1.

        :param resource_keys: A single value or a list-like or dict-like container specifying the
            entities of the resource to retrieve objects for, where the entity is given by the
            KeyField of the resource. If the resource_ids is a value or is list-like, then all
            objects corresponding to the entities are returned by default. If resource_ids is
            dict-like, then it is a mapping from the resource entity to an object_id_list.

            The object_id_list can take the values: '*', 0, or an array of positive ids. If it is
            '*', then all objects are returns. If it is 0, then the preferred object is returned.
            Otherwise, the ids will refer to the sequential index of the objects beginning with 1.

        :param media_types: A single or list-like container of acceptable media types for the
            server to return. If media_types is like-like, then the ordering specifies the
            preference of the media types to return, with the first being the most desirable. If
            the server is unable to provide the requested media type, it should return a 406
            Not Acceptable status, or if no objects exist for any media type then the server
            should return a 404 Not Found.

        :param location: Flag to indicate whether the object or a URL to the object should be
            returned. If location is set to True, it is up to the server to support this
            functionality and the lifetime of the returned URL is not given by the RETS
            specification.
        """
        headers = {
            'Accept': _build_accepted_media_types(media_types),
        }
        payload = {
            'Resource': resource,
            'Type': object_type,
            'ID': _build_entity_object_ids(resource_keys),
            'Location': int(location),
        }
        response = self._http_request(self._url_for('GetObject'), headers=headers, payload=payload)
        return parse_object(response)

    def _url_for(self, transaction: str) -> str:
        try:
            url = self._capabilities[transaction]
        except KeyError:
            raise RetsClientError('No URL found for transaction %s' % transaction)
        return urljoin(self._base_url, url)

    def _http_request(self, url: str, headers: dict = None, payload: dict = None) -> Response:
        if not self._session:
            raise RetsClientError('Session not instantiated. Call .login() first')

        request_headers = {
            **(headers or {}),
            'User-Agent': self.user_agent,
            'RETS-Version': self.rets_version,
            'RETS-UA-Authorization': self._rets_ua_authorization()
        }

        if self._use_get_method:
            if payload:
                url = '%s?%s' % (url, urlencode(payload))
            response = self._session.get(url, auth=self._http_auth, headers=request_headers)
        else:
            response = self._session.post(url, auth=self._http_auth, headers=request_headers, data=payload)

        response.raise_for_status()
        self._rets_session_id = self._session.cookies.get('RETS-Session-ID', '')
        return response

    def _rets_ua_authorization(self) -> str:
        return 'Digest ' + self._user_agent_auth_digest()

    def _user_agent_auth_digest(self) -> str:
        user_password = '%s:%s' % (self.user_agent, self._user_agent_password)
        a1 = md5(user_password.encode()).hexdigest()

        digest_values = '%s::%s:%s' % (a1, self._rets_session_id, self.rets_version)
        return md5(digest_values.encode()).hexdigest()


def _get_http_auth(username: str, password: str, auth_type: str) -> AuthBase:
    if auth_type == 'basic':
        return HTTPBasicAuth(username, password)
    if auth_type == 'digest':
        return HTTPDigestAuth(username, password)
    raise RetsClientError('unknown auth type %s' % auth_type)


def _build_entity_object_ids(entities: Union[str, Mapping[str, Any], Sequence[str]]) -> str:
    """
    Builds the string of object ids as required by the GetObject transaction request. See
    section 5.3 for the full definition:

    ID              ::= resource-set *(, resource-set)
    resource-set    ::= resource-entity [: object-id-list]
    resource-entity ::= 1*ALPHANUM
    object-id-list  ::= * | object-id *(: object-id)
    object-id       ::= 1*5DIGIT
    """
    if isinstance(entities, str):
        return _build_entity_object_ids((entities,))
    elif isinstance(entities, Sequence):
        return _build_entity_object_ids({e: '*' for e in entities})
    elif not isinstance(entities, Mapping):
        raise RetsClientError('Invalid entities argument')

    def _build_object_ids(object_ids: Any) -> str:
        if object_ids in ('*', 0, '0'):
            return str(object_ids)
        elif isinstance(object_ids, Sequence):
            return ':'.join(str(o) for o in object_ids)
        else:
            raise RetsClientError('Invalid entities argument')

    return ','.join('%s:%s' % (entity, _build_object_ids(object_ids))
                    for entity, object_ids in entities.items())


def _build_accepted_media_types(media_types: Union[str, Sequence[str]]) -> str:
    """
    Builds the Accept header of media types as required by the GetObject transaction request.
    The qvalue is used to specify the desirability of a given media type with 1 being the most
    desirable, 0 being the least, and a range in between.
    See section 5.1 for the full definition:

    Accept    ::= type / subtype [; parameter] *(, type / subtype [; parameter])
    type      ::= * | <publicly defined type>
    subtype   ::= * | <publicly defined subtype>
    parameter ::= q = <qvalue scale from 0 to 1>

    A complete list of media types is available at http://www.iana.org/assignments/media-types.
    """
    if isinstance(media_types, str):
        return media_types
    elif not isinstance(media_types, Sequence):
        raise RetsClientError('Invalid media types argument')

    n = float(len(media_types))
    return ','.join('%s;%.4f' % (types, 1 - i / n) for i, types in enumerate(media_types))



from typing import Any, Mapping, Sequence, Union

from rets.http import Object, RetsHttpClient


class ObjectType:

    def __init__(self, resource, metadata: dict, http_client: RetsHttpClient):
        self.resource = resource
        self._http = http_client
        self._metadata = metadata

    @property
    def name(self) -> str:
        return self._metadata['ObjectType']

    @property
    def mime_type(self) -> str:
        if 'MIMEType' in self._metadata:
            return self._metadata['MIMEType']
        return self._metadata['MimeType']

    @property
    def metadata(self) -> dict:
        return dict(self._metadata)

    def get(self, resource_keys: Union[str, Mapping[str, Any], Sequence[str]],
            **kwargs) -> Sequence[Object]:
        return self._http.get_object(self.resource.name, self.name, resource_keys, **kwargs)

    def __repr__(self) -> str:
        return '<Object: %s:%s>' % (self.resource.name, self.name)



from typing import Optional, Sequence

from rets.client.resource import Resource
from rets.client.utils import get_metadata_data
from rets.http import RetsHttpClient

"""
Example of metadata dict:

metadata = [{
    'ResourceID': 'Property',
    'KeyField': 'Matrix_Unique_ID',
    '_classes': [
        {
            'ClassName': 'Listing',
            'HasKeyIndex': '1',
            '_table': [
                ... column fields
            ],
        }
    ],
    '_object_types': [
        'ObjectType': 'LargePhoto',
        'MIMEType': 'image/jpeg',
    ]
}, {
    'ResourceID': 'Agent',
    'KeyField': 'Matrix_Unique_ID',
    '_classes': [
        {
            'ClassName': 'Listing',
            'HasKeyIndex': '1',
        }
    ],
}]
"""


class RetsClient:

    def __init__(self,
                 *args,
                 http_client: RetsHttpClient = None,
                 metadata: Sequence[dict] = (),
                 capability_urls: dict = None,
                 cookie_dict: dict = None,
                 **kwargs):
        self.http = http_client or RetsHttpClient(*args,
                                                  capability_urls=capability_urls, cookie_dict=cookie_dict,
                                                  **kwargs)
        if not (capability_urls and cookie_dict):
            self.http.login()
        self._resources = self._resources_from_metadata(metadata)

    @property
    def metadata(self) -> Sequence[dict]:
        return tuple(resource.metadata for resource in self._resources)

    @property
    def resources(self) -> Sequence[Resource]:
        if not self._resources:
            # TODO(ML) Differentiate between not having the metadata and
            # having an empty metadata
            self._resources = self._fetch_resources()
        return self._resources

    def get_resource(self, name: str) -> Optional[Resource]:
        for resource in self.resources:
            if resource.name == name:
                return resource
        raise KeyError('unknown resource %s' % name)

    def _fetch_resources(self) -> Sequence[Resource]:
        metadata = get_metadata_data(self.http, 'resource')
        return self._resources_from_metadata(metadata)

    def _resources_from_metadata(self, metadata: Sequence[dict]) -> Sequence[Resource]:
        return tuple(Resource(m, self.http) for m in metadata)



from typing import Optional, Sequence

from rets.client.resource_class import ResourceClass
from rets.client.object_type import ObjectType
from rets.client.utils import get_metadata_data
from rets.http import RetsHttpClient


class Resource:

    def __init__(self, metadata: dict, http_client: RetsHttpClient):
        self._http = http_client
        self._metadata = metadata
        self._classes = self._classes_from_metadata(metadata.get('_classes', ()))
        self._object_types = self._object_types_from_metadata(metadata.get('_object_types', ()))

    @property
    def name(self) -> str:
        return self._metadata['ResourceID']

    @property
    def key_field(self) -> str:
        return self._metadata['KeyField']

    @property
    def metadata(self) -> dict:
        metadata = dict(self._metadata)
        if self._classes:
            metadata['_classes'] = tuple(resource_class.metadata for resource_class in self._classes)
        if self._object_types:
            metadata['_object_types'] = tuple(object_type.metadata for object_type in self._object_types)
        return metadata

    @property
    def classes(self) -> Sequence[ResourceClass]:
        if not self._classes:
            self._classes = self._fetch_classes()
        return self._classes

    def get_class(self, name: str) -> Optional[ResourceClass]:
        for resource_class in self.classes:
            if resource_class.name == name:
                return resource_class
        raise KeyError('unknown class %s' % name)

    @property
    def object_types(self) -> Sequence[ObjectType]:
        if not self._object_types:
            self._object_types = self._fetch_object_types()
        return self._object_types

    def get_object_type(self, name: str) -> Optional[ObjectType]:
        for resource_object in self.object_types:
            if resource_object.name == name:
                return resource_object
        raise KeyError('unknown object type %s' % name)

    def _fetch_classes(self) -> Sequence[ResourceClass]:
        metadata = get_metadata_data(self._http, 'class', resource=self.name)
        return self._classes_from_metadata(metadata)

    def _fetch_object_types(self) -> Sequence[ObjectType]:
        metadata = get_metadata_data(self._http, 'object', resource=self.name)
        return self._object_types_from_metadata(metadata)

    def _classes_from_metadata(self, classes_metadata: Sequence[dict]) -> Sequence[ResourceClass]:
        return tuple(ResourceClass(self, m, self._http) for m in classes_metadata)

    def _object_types_from_metadata(self, object_types_metadata: Sequence[dict]) -> Sequence[ObjectType]:
        return tuple(ObjectType(self, m, self._http) for m in object_types_metadata)

    def __repr__(self) -> str:
        return '<Resource: %s>' % self.name



from typing import Sequence

from rets.http import Object


class Record:

    def __init__(self, resource_class, data: dict):
        self.resource = resource_class.resource
        self.resource_class = resource_class
        self.resource_key = str(data[resource_class.resource.key_field])
        self.data = data

    def get_objects(self, name: str, **kwargs) -> Sequence[Object]:
        resource_object = self.resource.get_object_type(name)
        return resource_object.get(self.resource_key, **kwargs)

    def __repr__(self) -> str:
        return '<Record: %s:%s:%s>' % (
            self.resource_class.resource.name,
            self.resource_class.name,
            self.resource_key,
        )



import sys

from setuptools import setup

if sys.version_info < (3, 5):
    print('rets requires Python 3.5 or later')
    sys.exit(1)


long_desc = 'Python 3 client for the Real Estate Transaction Standard (RETS) Version 1.7.2'

install_requires = [
    'requests>=2.12.3',
    'requests-toolbelt>=0.7.0,!=0.9.0',
    'udatetime==0.0.16',
    'docopts',
    'lxml>=4.3.0',
]

setup_requires = [
    'pytest-runner',
]

tests_requires = [
    'flake8',
    'pytest',
]

packages = [
    'rets',
    'rets.client',
    'rets.http',
    'rets.http.parsers',
]

setup(
    name='rets-python',
    version='0.4.5',
    description='rets-python',
    long_description=long_desc,
    author='Martin Liu',
    author_email='martin@opendoor.com',
    url='https://github.com/opendoor-labs/rets',
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Financial and Insurance Industry',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
    ],
    license='MIT License',
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_requires,
    packages=packages,
)




class RetsError(RuntimeError):
    pass


class RetsClientError(RetsError):
    pass


class RetsParseError(RetsClientError):
    pass


class RetsResponseError(RetsClientError):

    def __init__(self, content: str, headers: dict):
        super().__init__('Unexpected response from RETS')
        self.content = content
        self.headers = headers


class RetsApiError(RetsClientError):

    def __init__(self, reply_code: int, reply_text: str, xml: str):
        super().__init__('[%i] %s\n\n%s' % (reply_code, reply_text, xml))
        self.reply_code = reply_code
        self.reply_text = reply_text
        self.xml = xml



import logging
import re
from collections import OrderedDict
from datetime import datetime, time, timezone
from decimal import Decimal
from functools import partial
from typing import Any, Sequence

import udatetime

from rets.errors import RetsParseError

logger = logging.getLogger('rets')


class RecordDecoder:

    def __init__(self, table: Sequence[dict], include_tz: bool = False):
        self._metadata_map = {field['SystemName']: field for field in table}
        self._include_tz = include_tz

    def decode(self, rows: Sequence[dict]) -> Sequence[dict]:
        if not rows:
            return ()

        # Build dict of field to decoder functions, assuming that all rows have the same fields.
        decoders = self._build_decoders(tuple(rows[0].keys()))

        def decode_field(field: str, value: str) -> Any:
            if value == '':
                return None
            return decoders[field](value)

        return tuple(OrderedDict((field, decode_field(field, value)) for field, value in row.items())
                     for row in rows)

    def _build_decoders(self, fields: Sequence[str]) -> dict:
        decoders = {}
        for field in fields:
            try:
                field_metadata = self._metadata_map[field]
            except KeyError:
                logger.warning('field %s not found in table metadata', field)
                field_metadata = {'DataType': 'Character'}

            decoders[field] = _get_decoder(
                data_type=field_metadata['DataType'],
                interpretation=field_metadata.get('Interpretation', ''),
                include_tz=self._include_tz,
            )

        return decoders


def _get_decoder(data_type: str, interpretation: str, include_tz: bool = False):
    if interpretation == _LOOKUP_TYPE:
        return str
    elif interpretation in _LOOKUP_MULTI_TYPES:
        return lambda value: value.split(',')

    if data_type in _TIMEZONE_AWARE_DECODERS:
        return partial(_TIMEZONE_AWARE_DECODERS[data_type], include_tz=include_tz)

    try:
        return _DECODERS[data_type]
    except KeyError:
        raise RetsParseError('unknown data type %s' % data_type) from None


def _decode_datetime(value: str, include_tz: bool) -> datetime:
    # Correct `0000-00-00 00:00:00` to `0000-00-00T00:00:00`
    if re.match(r'^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}$', value):
        value = '%sT%s' % (value[0:10], value[11:])
    # Correct `0000-00-00` to `0000-00-00T00:00:00`
    elif re.match(r'^\d{4}-\d{2}-\d{2}$', value):
        value = '%sT00:00:00' % value[0:10]

    decoded = udatetime.from_string(value)
    if not include_tz:
        return decoded.astimezone(timezone.utc).replace(tzinfo=None)
    return decoded


def _decode_time(value: str, include_tz: bool) -> time:
    decoded = _decode_datetime('1970-01-01T' + value, include_tz)
    return decoded.time().replace(tzinfo=decoded.tzinfo)


def _decode_date(value: str, include_tz: bool) -> datetime:
    try:
        decoded = datetime.strptime(value, '%Y-%m-%d')
        return decoded
    except ValueError:
        return _decode_datetime(value, include_tz)


_LOOKUP_TYPE = 'Lookup'

_LOOKUP_MULTI_TYPES = frozenset(('LookupMulti', 'LookupBitstring', 'LookupBitmask'))

_TIMEZONE_AWARE_DECODERS = {
    'DateTime': _decode_datetime,
    'Time': _decode_time,
    'Date': _decode_date,
}

_DECODERS = {
    'Boolean': lambda value: value == '1',
    'Character': str,
    'Tiny': int,
    'Small': int,
    'Int': int,
    'Long': int,
    'Decimal': Decimal,
    'Number': int,
    # Point is new "Edm.GeographyPoint" from RESO, look online for spec. Can store as Postgres Point, see https://bit.ly/2BDPgUS
    'Point': str,  
}
